#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See the README.md in the repository for more info
#
"""packet wide constants"""

__version__ = "0.17.3"

XDG_CONFIG_HOME = "XDG_CONFIG_HOME"
